# -*- coding: utf-8 -*-
# -*- mode: python -*-
"""Common functions for plotting things. 

Functions in this module should take matplotlib axes as their first argument,
but there is no direct import of matplotlib.

"""
from typing import Tuple

import numpy as np


def spectrogram(
    ax,
    signal: np.ndarray,
    sampling_rate_hz: float,
    *,
    window_s: float = 0.020,
    shift_s: float = 0.010,
    frequency_range: Tuple[float, float] = (700, 10000),
    compression: float = 0.1,
    **plot_kwargs
):
    """Plot a spectrogram of a signal, with some useful defaults for birdsong"""
    from math import ceil, log

    from ewave import rescale
    from libtfr import fgrid, mfft_precalc, tgrid

    data = rescale(signal, np.float64)
    window_samples = int(window_s * sampling_rate_hz)
    nfft = 2 ** ceil(log(window_samples, 2))
    shift_samples = int(shift_s * sampling_rate_hz)
    mfft = mfft_precalc(nfft, np.hanning(window_samples))
    spec = mfft.mtspec(data, shift_samples)
    f, fi = fgrid(sampling_rate_hz, nfft, frequency_range)
    t = tgrid(spec, sampling_rate_hz, shift_samples)
    ax.imshow(
        np.log(spec[fi, :] + compression) - np.log(compression),
        aspect="auto",
        origin="lower",
        extent=(t[0], t[-1], f[0], f[-1]),
        **plot_kwargs
    )


def simple_axes(*axes):
    """Simple axes: only bottom and right lines shown"""
    for ax in axes:
        ax.spines["top"].set_visible(False)
        ax.spines["right"].set_visible(False)
        ax.get_xaxis().tick_bottom()
        ax.get_yaxis().tick_left()


def hide_axes(*axes):
    """Hidden axes: no axis lines shown"""
    for ax in axes:
        ax.get_xaxis().set_visible(False)
        ax.get_yaxis().set_visible(False)
        ax.set_frame_on(False)


def adjust_raster_ticks(ax, gap=0):
    """Adjust raster marks to have gap pixels between them (sort of)"""
    miny, maxy = ax.get_ylim()
    ht = ax.get_window_extent().height
    for p in ax.lines:
        p.set_markersize(ht / ((maxy - miny)) - gap)



# -*- coding: utf-8 -*-
# -*- mode: python -*-
""" Functions for processing spike trains (point processes) """
from typing import Optional, Tuple, Any
from pathlib import Path
from dataclasses import dataclass

import h5py as h5
import numpy as np
from numpy.typing import ArrayLike


def psth(
    spikes: ArrayLike,
    binwidth: float,
    *,
    start: Optional[float] = None,
    stop: Optional[float] = None,
) -> Tuple[np.ndarray, np.ndarray]:
    """Compute a PSTH of spike times.

    spikes: array or list of spike times
    binwidth: the bin duration (same units as spikes)
    start: the start of the observation interval. If None, the time of the first spike is used.
    stop: the end of the observation interval. If None, the time of the last spike plus one bin is used.

    Returns (spike counts, bin times)
    """
    spikes = np.asarray(spikes)
    t1 = start if start is not None else spikes.min()
    t2 = stop if stop is not None else (spikes.max() + binwidth)
    bins = np.arange(t1, t2, binwidth)
    counts, bins = np.histogram(spikes, bins)
    return counts, bins[:-1]


def rate(
    spikes: ArrayLike,
    binwidth: float,
    kernel: ArrayLike,
    *,
    start: Optional[float] = None,
    stop: Optional[float] = None,
) -> Tuple[np.ndarray, np.ndarray]:
    """Compute a smoothed estimate of spike rate from spike times.

    spikes: array or list of spike times
    binwidth: the bin duration (same units as spikes)
    kernel: the smoothing kernel. Sampling interval needs to be same as binwidth.
    start: the start of the observation interval. If None, the time of the first spike is used.
    stop: the end of the observation interval. If None, the time of the last spike is used.

    Returns (rate estimate, bin times)
    """
    counts, bins = psth(spikes, binwidth, start=start, stop=stop)
    return np.convolve(counts, kernel, mode="same"), bins


@dataclass
class SpikeWaveforms:
    """
    waveform: nspikes x npoints array
    times: nspikes array (times of spikes in units of samples)
    sampling_rate: the sampling rate of the spikes and the spike times (in Hz)
    peak_index: the index corresponding to the time of the spike in the waveform
    """

    waveforms: ArrayLike
    times: ArrayLike
    sampling_rate: int
    peak_index: int


def save_waveforms(
    path: Path,
    waveforms: SpikeWaveforms,
    **attributes: Any,
) -> None:
    """Save spike waveforms to an hdf5 file

    path: the location of the file (will overwrite)
    attributes: any additional metadata to store in the file

    Waveforms should be stored as they were recorded. The `peak_index` argument
    will typically refer to the maximum negative deflection of the spike given
    typical recording parameters.

    """
    spikes = np.asarray(waveforms.waveforms)
    times = np.asarray(waveforms.times)
    ntimes, nspikes = spikes.shape
    if ntimes != times.size:
        raise ValueError(
            "number of rows in waveform array must match number of elements in times array"
        )
    with h5.File(path, "w") as fp:
        # unchunked storage uses the most space but allows random access
        dset_spikes = fp.create_dataset("waveforms", data=spikes)
        dset_spikes.attrs["sampling_rate"] = waveforms.sampling_rate
        dset_spikes.attrs["peak_index"] = waveforms.peak_index
        dset_times = fp.create_dataset("times", data=times)
        dset_times.attrs["sampling_rate"] = waveforms.sampling_rate

        fp.attrs["schema"] = "meliza-org.spikewaveforms"
        fp.attrs["schema_ver"] = 1
        for k, v in attributes.items():
            fp.attrs[k] = v


def load_waveforms(path: Path) -> SpikeWaveforms:
    """Load spike waveforms from an hdf5 file.

    Data are returned in a SpikeWaveforms structure, plus the top-level
    attributes in the "attrs" property. The data are loaded lazily, so the file
    handle remains open.

    """
    fp = h5.File(path, "r")
    dset_spikes = fp["waveforms"]
    dset_times = fp["times"]
    out = SpikeWaveforms(
        dset_spikes,
        dset_times,
        dset_spikes.attrs["sampling_rate"],
        dset_spikes.attrs["peak_index"],
    )
    out.attrs = dict(fp.attrs)
    return out



# -*- coding: utf-8 -*-
# -*- mode: python -*-
""" Functions for using kilosort/phy data """
import json
import logging
import re
from collections import namedtuple
from functools import lru_cache
from pathlib import Path
from typing import Dict, Iterator

import anyio
import ewave
import h5py as h5
import numpy as np
import pandas as pd
import quickspikes as qs
import toelis
from httpx import AsyncClient

from dlab import nbank, pprox
from dlab.spikes import SpikeWaveforms, save_waveforms

logging.getLogger(__name__).addHandler(logging.NullHandler())

Trial = namedtuple(
    "Trial",
    [
        "recording_entry",
        "recording_start",
        "recording_end",
        "stimulus_name",
        "stimulus_start",
        "stimulus_end",
    ],
)

Stimulus = namedtuple("Stimulus", ["name", "start", "end"], defaults=[None])


def read_kilo_params(fname: Path) -> Dict:
    """Read the kilosort params.py file"""
    from configparser import ConfigParser
    from itertools import chain

    parser = ConfigParser()
    with open(fname, "rt") as lines:
        lines = chain(("[top]",), lines)
        parser.read_file(lines)
    sect = parser["top"]
    return dict(
        dtype=sect["dtype"].strip("'"),
        nchannels=int(sect["n_channels_dat"]),
        sampling_rate=float(sect["sample_rate"]),
    )


def oeaudio_stims(dset: h5.Dataset) -> Iterator[Stimulus]:
    """Parse the messages in the 'stim' dataset to get a table of stimuli with
    start samples. Note that these need to be corrected for offset of the
    recording and network lag.

    """
    re_start = re.compile(r"start (.*)")
    for row in dset:
        time = row["start"]
        message = row["message"].decode("utf-8")
        m = re_start.match(message)
        if m is not None:
            stim_name = Path(m.group(1)).stem
            yield Stimulus(stim_name, time)


async def stim_duration(session: AsyncClient, stim_name: str) -> float:
    """
    Returns the duration of a stimulus (in s). This can only really be done by
    downloading the stimulus from the registry, because the start/stop times are
    not reliable. We try to speed this up by memoizing the function and caching
    the downloaded files.

    """
    target = await nbank.find_resource(session, nbank.default_registry, stim_name)
    with ewave.wavfile(str(target)) as fp:
        return 1.0 * fp.nframes / fp.sampling_rate


async def oeaudio_to_trials(
    session: AsyncClient,
    data_file: h5.File,
    sync_dset: str,
    sync_thresh: float = 1.0,
    prepad: float = 1.0,
) -> Iterator[Trial]:
    """Extracts trial information from an oeaudio-present experiment ARF file

    When using oeaudio-present, a single recording is made in response to all
    the stimuli. The stimulus presentation script sends network events to
    open-ephys to mark the start and stop of each stimulus. There is typically a
    significant lag between the 'start' event and the onset of the stimulus, due
    to buffering of the audio playback. However, the oeaudio-present script will
    play a synchronization click on a second channel by default. As long as the
    user remembers to record this channel, it can be used to correct the onset
    and offset values.

    The continuous recording is broken up into trials based on the stimulus
    presentation, such that each trial encompasses one and only one stimulus.
    The `prepad` parameter specifies, in seconds, when trials begin relative to
    stimulus onset. The default is 1.0 s.

    """
    from itertools import zip_longest

    from dlab.extracellular import (
        entry_datetime,
        entry_time,
        find_stim_dset,
        iter_entries,
    )

    expt_start = None
    det = qs.detector(sync_thresh, 10)
    trials = []
    for entry_num, entry in iter_entries(data_file):
        logging.info(" - entry: '%s'", entry.name)
        entry_start = entry_time(entry)
        logging.info("  - start time: %s", entry_datetime(entry))
        if expt_start is None:
            expt_start = entry_start

        logging.info("  - parsing stimulus log")
        stims = list(oeaudio_stims(find_stim_dset(entry)))
        logging.info("    - detected %d stimuli", len(stims))
        logging.info("  - sync track: '%s'", sync_dset)
        sync = entry[sync_dset]
        sync_data = sync[:].astype("d")
        det.scale_thresh(sync_data.mean(), sync_data.std())
        stim_onsets = np.asarray(det(sync_data))
        logging.info("    - detected %d clicks", stim_onsets.size)
        dset_offset = sync.attrs["offset"]
        dset_end = sync.size
        sampling_rate = sync.attrs["sampling_rate"]
        stim_sample_offset = int(dset_offset * sampling_rate)
        logging.info("  - recording clock offset: %d", stim_sample_offset)

        if len(stims) != stim_onsets.size:
            logging.warning(
                "  - WARNING: number of stimuli does not match number of clicks. This recording may need to be discarded"
            )

        padding_samples = int(prepad * sampling_rate)
        for stim, onset, offset in zip_longest(
            stims, stim_onsets, stim_onsets[1:], fillvalue=dset_end + padding_samples
        ):
            stim_seconds = await stim_duration(session, stim.name)
            stim_samples = int(stim_seconds * sampling_rate)
            if stim_samples > offset - onset:
                logging.warning(
                    "  - WARNING: stimulus %s is longer than the duration of the trial",
                    stim,
                )
            trials.append(
                Trial(
                    entry_num,
                    onset - padding_samples,
                    offset - padding_samples,
                    stim.name,
                    onset,
                    onset + stim_samples,
                )
            )
    return trials


def assign_events_flat(events: pd.DataFrame, sampling_rate: float):
    """Assign event_times to clusters, generating a large toelis object"""
    nevents, _ = events.shape
    nclusters = events.index.unique().size
    logging.info("- grouping %d spikes into %d clusters...", nevents, nclusters)
    return events.groupby("clust").apply(
        lambda df: df.time.sort_values().to_numpy() / sampling_rate * 1000.0
    )


def trials_to_pprox(trials: pd.DataFrame, sampling_rate: float):
    """Convert pandas trials to pproc"""
    for trial in trials.itertuples():
        if isinstance(trial.events, float):
            events = []
        else:
            events = (trial.events.astype("d") - trial.stimulus_start) / sampling_rate
        pproc = {
            "events": events,
            "offset": trial.stimulus_start / sampling_rate,
            "index": trial.Index,
            "interval": (
                (trial.recording_start - trial.stimulus_start) / sampling_rate,
                (trial.recording_end - trial.stimulus_start) / sampling_rate,
            ),
            "stimulus": {
                "name": trial.stimulus_name,
                "interval": (
                    0.0,
                    (trial.stimulus_end - trial.stimulus_start) / sampling_rate,
                ),
            },
            "recording": {
                "entry": trial.recording_entry,
                "start": trial.recording_start,
                "end": trial.recording_end,
            },
        }
        yield pproc


async def group_spikes_script(argv=None):
    import argparse
    import os

    from dlab.core import __version__
    from dlab.extracellular import entry_metadata, iter_entries
    from dlab.util import json_serializable

    version = "2023.05.11"

    p = argparse.ArgumentParser(
        description="group kilosorted spikes into pprox files based on cluster and trial"
    )
    p.add_argument(
        "-v",
        "--version",
        action="version",
        version=f"%(prog)s {version} (melizalab-tools {__version__})",
    )
    p.add_argument("--debug", help="show verbose log messages", action="store_true")
    p.add_argument(
        "--dry-run",
        help="do everything except write the output files",
        action="store_true",
    )
    p.add_argument(
        "--sync",
        default="sync",
        help="name of channel with synchronization signal (default '%(default)s')",
    )
    p.add_argument(
        "--sync-thresh",
        default=30.0,
        type=float,
        help="threshold (z-score) for detecting sync clicks (default %(default)0.1f)",
    )
    p.add_argument(
        "--prepad",
        type=float,
        default=1.0,
        help="sets trial start time relative to stimulus onset (default %(default)0.1f s)",
    )
    p.add_argument(
        "--toelis",
        action="store_true",
        help="output toelis instead of pprox. one file will be generated for "
        "the entire recording (including multiunits)",
    )
    p.add_argument(
        "--cluster",
        "-c",
        help="only save data for the specified clusters (as comma-separated list)",
        type=lambda s: [int(item) for item in s.split(",")],
    )
    p.add_argument(
        "--output",
        "-o",
        type=Path,
        default=".",
        help="directory to output pprox files (default current directory)",
    )
    p.add_argument(
        "--mua",
        action="store_true",
        help="save multiunit clusters along with single units",
    )
    p.add_argument(
        "--artifact-reject-thresh",
        type=float,
        default=6.0,
        help="threshold for rejecting artifact spikes (max absolute amplitude"
        " more than x times max absolute amplitude of the mean spike)",
    )
    p.add_argument(
        "--no-waveforms",
        "-W",
        action="store_true",
        help="if set, do not save representative waveforms from each unit's main channel in an hdf5 file",
    )
    p.add_argument(
        "--waveform-pre-peak",
        type=float,
        default=2.0,
        help="samples before the spike to keep (default %(default).1f ms)",
    )
    p.add_argument(
        "--waveform-post-peak",
        type=float,
        default=5.0,
        help="samples after the spike to keep (default %(default).1f ms)",
    )
    p.add_argument("recording", type=Path, help="path of ARF recording file")
    p.add_argument(
        "sortdir",
        type=Path,
        help="kilosort output directory. Needs to contain 'spike_times.npy', 'spike_clusters.npy',"
        " 'cluster_info.tsv', and 'temp_wh.dat'",
    )
    args = p.parse_args(argv)
    logging.basicConfig(
        format="%(message)s", level=logging.DEBUG if args.debug else logging.INFO
    )
    os.environ["HDF5_USE_FILE_LOCKING"] = "FALSE"
    logging.info("- %s version %s", p.prog, version)

    async with AsyncClient(timeout=None) as session:
        resource_info = await nbank.fetch_metadata(
            session, nbank.default_registry, args.recording.stem
        )
        recording_name = resource_info["name"]
        resource_url = nbank.registry.full_url(nbank.default_registry, recording_name)
        logging.info("- kilosort output directory: %s", args.sortdir)
        timefile = args.sortdir / "spike_times.npy"
        clustfile = args.sortdir / "spike_clusters.npy"
        infofile = args.sortdir / "cluster_info.tsv"
        logging.info("  - spike times: %s", timefile)
        logging.info("  - spike clusters: %s", clustfile)
        events = pd.DataFrame(
            {"time": np.load(timefile).squeeze(), "clust": np.load(clustfile)},
        ).set_index("clust")
        logging.info("  - cluster info: %s", infofile)
        info = pd.read_csv(infofile, sep="\t", index_col=0)
        recfile = args.sortdir / "temp_wh.dat"
        params = read_kilo_params(args.sortdir / "params.py")
        recording = np.memmap(recfile, mode="c", dtype=params["dtype"])
        recording = np.reshape(
            recording, (recording.size // params["nchannels"], params["nchannels"])
        )
        nsamples, nchannels = recording.shape
        logging.info("  - filtered recording: %s", recfile)
        logging.info("    - %d samples, %d channels", nsamples, nchannels)
        if args.cluster is not None:
            logging.info("- only analyzing clusters: %s", args.cluster)
            events = events.loc[args.cluster]

        if args.toelis:
            clusters = assign_events_flat(events, params["sampling_rate"])
            outfile = (args.output / recording_name).with_suffix(".toe_lis")
            if not args.dry_run:
                with open(outfile, "wt") as ofp:
                    toelis.write(ofp, clusters)
                logging.info(
                    "- saved %d spikes to '%s'", toelis.count(clusters), outfile
                )
            return

        if args.recording.is_file():
            datafile = args.recording
        else:
            datafile = await nbank.find_resource(
                session, nbank.default_registry, str(args.recording)
            )
        logging.info("- splitting '%s' into trials:", datafile)
        with h5.File(datafile, "r") as afp:
            trials = pd.DataFrame(
                await oeaudio_to_trials(
                    session, afp, args.sync, args.sync_thresh, args.prepad
                )
            )
            entry_attrs = tuple(entry_metadata(e) for _, e in iter_entries(afp))

        # this pandas magic sorts the events by cluster and trial
        logging.info("- sorting events into trials:")
        events["trial"] = (
            trials.recording_start.searchsorted(events.time, side="left") - 1
        )

        total_spikes = 0
        total_clusters = 0
        for clust_id, cluster in events.groupby("clust"):
            clust_info = info.loc[clust_id]
            clust_type = clust_info["group"]
            n_spikes = len(cluster)
            if clust_type == "noise" or (clust_type == "mua" and not args.mua):
                logging.info(
                    "  - cluster %d (%d spikes, %s) -> skipped",
                    clust_id,
                    n_spikes,
                    clust_type,
                )
                continue
            logging.info(
                "  ✓ cluster %d (%d spikes, %s)",
                clust_id,
                n_spikes,
                clust_type,
            )
            # remove artifact spikes
            n_before = int(args.waveform_pre_peak * params["sampling_rate"] / 1000)
            n_after = int(args.waveform_post_peak * params["sampling_rate"] / 1000)
            spikes = cluster[
                (cluster.time > n_before) & (cluster.time < (nsamples - n_after))
            ]
            waveforms = qs.peaks(
                recording[:, clust_info["ch"]],
                spikes.time,
                n_before=n_before,
                n_after=n_after,
            )
            mean_spike = waveforms.mean(0)
            included = np.abs(waveforms).max(-1) < (
                np.abs(mean_spike).max(-1) * args.artifact_reject_thresh
            )
            n_included = included.sum()
            if n_included < n_spikes:
                cluster = cluster[included]
                waveforms = waveforms[included]
                logging.info(
                    "    - %d artifact spike(s) excluded", n_spikes - n_included
                )
            # aggregate spikes by trial and left join to trial information table
            # - empty trials will be nan
            clust_trials = trials.join(
                cluster.groupby("trial")
                .apply(lambda x: x.time.to_numpy())
                .rename("events")
            )
            total_spikes += n_spikes
            total_clusters += 1
            outfile = args.output / f"{recording_name}_c{clust_id}.pprox"
            logging.info(
                "    - %d spikes -> %s",
                n_included,
                outfile,
            )
            clust_trials = pprox.from_trials(
                trials_to_pprox(clust_trials, params["sampling_rate"]),
                schema=pprox._stimtrial_schema,
                recording=resource_url,
                processed_by=[f"{p.prog} {version}"],
                kilosort_amplitude=clust_info["Amplitude"],
                kilosort_contam_pct=clust_info["ContamPct"],
                kilosort_source_channel=clust_info["ch"],
                kilosort_probe_depth=clust_info["depth"],
                kilosort_n_spikes=clust_info["n_spikes"],
                entry_metadata=entry_attrs,
                **resource_info["metadata"],
            )
            if not args.dry_run:
                with open(outfile, "wt") as ofp:
                    json.dump(clust_trials, ofp, default=json_serializable)
                if not args.no_waveforms:
                    outfile = args.output / (outfile.stem + "_spikes.h5")
                    logging.info(
                        "    - waveforms on channel %d -> %s",
                        clust_info["ch"],
                        outfile,
                    )
                    save_waveforms(
                        outfile,
                        SpikeWaveforms(
                            waveforms,
                            cluster.time.to_numpy(),
                            params["sampling_rate"],
                            n_before,
                        ),
                        recording=resource_url,
                        processed_by=f"{p.prog} {version}",
                        kilosort_amplitude=clust_info["Amplitude"],
                        kilosort_contam_pct=clust_info["ContamPct"],
                        kilosort_source_channel=clust_info["ch"],
                        kilosort_probe_depth=clust_info["depth"],
                        kilosort_n_spikes=clust_info["n_spikes"],
                    )

        logging.info(
            "- a total of %d spikes were assigned to %d clusters",
            total_spikes,
            total_clusters,
        )


def run_group_spikes():
    anyio.run(group_spikes_script)



# -*- coding: utf-8 -*-
# -*- mode: python -*-
""" Commandline script to extract waveforms from extracellular recordings """
import json
import logging

import h5py as h5
import numpy as np
import pandas as pd

import nbank


if __name__ == "__main__":
    import argparse

    from dlab.core import __version__, get_or_verify_datafile

    script_version = "2022.06.24"

    p = argparse.ArgumentParser(
        description="extract waveforms from extracellular recording based "
        "on spike times in pprox files. A csv file is generated"
        "for each pprox file"
    )
    p.add_argument(
        "-v",
        "--version",
        action="version",
        version=f"%(prog)s {script_version} (melizalab-tools {__version__})",
    )
    p.add_argument("--debug", help="show verbose log messages", action="store_true")
    p.add_argument(
        "--num-spikes",
        "-n",
        type=int,
        default=100,
        help="maximum number of spikes to extract",
    )
    p.add_argument(
        "--pre-spike",
        type=float,
        default=2.0,
        help="samples before the spike to keep (default %(default).1f ms)",
    )
    p.add_argument(
        "--post-spike",
        type=float,
        default=5.0,
        help="samples after the spike to keep (default %(default).1f ms)",
    )
    p.add_argument(
        "--upsample",
        type=float,
        default=3,
        help="factor to upsample spikes before aligning them (default %(default)0.1f)",
    )
    p.add_argument(
        "--seed", type=int, default=12345, help="random seed for selecting spikes"
    )
    p.add_argument("recording", help="ARF recording (local file, or neurobank id/URL)")
    p.add_argument(
        "pprox",
        nargs="+",
        type=argparse.FileType("r", encoding="utf-8"),
        help="pprox files with spike times",
    )
    args = p.parse_args()
    logging.basicConfig(
        format="%(message)s", level=logging.DEBUG if args.debug else logging.INFO
    )

    datafile, arf_info = get_or_verify_datafile(args.recording, args.debug)

    with h5.File(datafile, "r") as arf:
        for spikefile in args.pprox:
            logging.info(" - processing spike times in '%s':", spikefile.name)
            spike_data = json.load(spikefile)
            # check that the recording matches the ARF file
            logging.info("   - source recording: %s", spike_data["recording"])
            dset_channel = spike_data["kilosort_source_channel"]
            # TODO verify that that index is 1-based
            dset_name = f"CH{dset_channel + 1}"
            logging.info("   - main channel: %s", dset_name)
            resource_info = nbank.describe(spike_data["recording"])
            if resource_info["sha1"] != arf_info["sha1"]:
                logging.warning(
                    "   - warning: recording does not match the supplied ARF file, skipping"
                )
                continue
            # convert spike times to sample counts (relative to entry start)
            entry_metadata = spike_data["entry_metadata"]
            if len(set(e["sampling_rate"] for e in entry_metadata)) > 1:
                raise ValueError("Not all entries have the same sampling rate")
            sampling_rate = entry_metadata[0]["sampling_rate"]
            dsets = [arf[m["name"]][dset_name] for m in entry_metadata]
            n_before = int(args.pre_spike * sampling_rate / 1000)
            n_after = int(args.post_spike * sampling_rate / 1000)
            spikes = []
            for trial in spike_data["pprox"]:
                entry = trial["recording"]["entry"]
                sampling_rate = entry_metadata[entry]["sampling_rate"]
                for spike in trial["events"]:
                    peak_sample = (
                        int(spike * sampling_rate) + trial["recording"]["start"]
                    )
                    spike_start = peak_sample - n_before
                    spike_end = peak_sample + n_after
                    if spike_start > 0 and spike_end < trial["recording"]["stop"]:
                        spikes.append((entry, spike, spike_start, spike_end))
            spikes = pd.DataFrame.from_records(
                np.array(
                    spikes,
                    dtype=[
                        ("entry", "i4"),
                        ("peak", "f8"),
                        ("start", "i8"),
                        ("stop", "i8"),
                    ],
                )
            )
            nspikes, _ = spikes.shape
            selected = spikes.sample(
                min(args.num_spikes, nspikes), random_state=args.seed
            )
            nspikes, _ = selected.shape

            waveforms = np.zeros((nspikes, n_before + n_after), dtype="f8")
            for i, row in enumerate(selected.itertuples(index=False)):
                waveforms[i, :] = dsets[row.entry][row.start : row.stop]
            # __import__("IPython").embed()

            #     spike_idx = (np.asarray(trial["events"]) * sampling_rate).astype("int64") + trial["recording"]["start"]
            #     nspikes = spike_idx.size
            #     logging.debug("    - trial %d: %d spikes from entry %d", trial["index"], nspikes, entry)
            #     spike_times.append(spike_idx)
            #     spike_entries.append(np.ones(nspikes, dtype="int32") * entry)
            # all_spike_entries = np.concatenate(spike_entries)
            # all_spike_times = np.concatenate(spike_times)
            # nspikes = all_spike_times.size

            # TODO sort by entry to avoid looking up the dataset on every spike
            # spike_tbl = pd.DataFrame({"entry":
            # spikes = []
            # for entry, peak in zip(all_spike_entries[selected], all_spike_times[selected]):
            #
            #     s_before = peak - args.pre_spike * entry_metadata[entry]["sampling_rate"]
            #     s_after = peak + args.post_spike * entry_metadata[entry]["sampling_rate"]
            #     spikes.append(dset[s_before:s_after])



# -*- coding: utf-8 -*-
# -*- mode: python -*-
import pytest
from dlab import signal

binsize = 0.01
bandwidth = 0.1
kernels = [
    "gaussian",
    "exponential",
    "biweight",
    "triweight",
    "cosinus",
    "epanech",
    "hanning",
]


@pytest.fixture(params=kernels)
def conv_kernel(request):
    return signal.kernel(request.param, bandwidth, binsize)


def test_kernel_scale(conv_kernel):
    k, kt = conv_kernel
    assert k.size == kt.size
    assert k.sum() * binsize == pytest.approx(1.0)



# -*- coding: utf-8 -*-
# -*- mode: python -*-
import pytest
from dlab import spikes


def test_psth_empty():
    events = []
    with pytest.raises(ValueError):
        _ = spikes.psth(events, 0.01)
    counts, bins = spikes.psth(events, 0.01, start=0.0, stop=1.0)
    assert counts.sum() == 0
    assert counts.size == bins.size


def test_psth_auto_bins():
    binwidth = 0.01
    events = [1.1, 2.1, 2.9, 4.01]
    counts, bins = spikes.psth(events, binwidth)
    assert counts.sum() == len(events)
    assert counts.size == bins.size
    assert bins[0] == events[0]
    assert bins[-1] + binwidth > events[-1]


def test_psth_clip():
    start = 1.1
    stop = 4.0
    binwidth = 0.01
    events = [1.1, 2.1, 2.9, 4.01]
    counts, bins = spikes.psth(events, binwidth, start=start, stop=stop)
    assert counts.sum() == sum(1 for e in events if e >= start and e < 4.0)
    assert counts.size == bins.size


def test_psth_multi_trial():
    """multi-trial data is fine as long as it can be flattened"""
    binwidth = 0.1
    events = [[1.1, 2.1, 2.9, 4.01], [1.0, 2.2, 3.0, 4.1]]
    counts, bins = spikes.psth(events, binwidth)
    assert counts.sum() == 8
    assert counts.size == bins.size


def test_rate_scale():
    """If kernel is scaled correctly, sum of rate over interval should be equal to N"""
    from dlab.signal import kernel

    binwidth = 0.01
    bandwidth = 0.1
    events = (1.1, 2.1, 2.9, 4.01)
    k, kt = kernel("gaussian", bandwidth, binwidth)
    r, rt = spikes.rate(events, binwidth, k, start=0.0, stop=5.0)
    assert r.sum() * binwidth == pytest.approx(len(events))



# -*- coding: utf-8 -*-
# -*- mode: python -*-
"""Process pprox (point process) data

See https://meliza.org/spec:2/pprox/ for specification

This package deliberately does not include any functions for serialization, as
pprox objects are just python dictionaries.

Copyright (C) Dan Meliza, 2006-2020 (dan@meliza.org)

"""
from typing import Iterable, Dict

_base_schema = "https://meliza.org/spec:2/pprox.json#"
_stimtrial_schema = "https://meliza.org/spec:2/stimtrial.json#"


def empty():
    """Returns a new, empty pprox object"""
    return from_trials([])


def from_trials(trials: Iterable, *, schema: str = _base_schema, **metadata) -> Dict:
    """Wrap a sequence of trials in a pprox object, optionally specifying top-level metadata"""
    d = {"$schema": schema, "pprox": tuple(trials)}
    d.update(**metadata)
    return d


def wrap_uuid(b):
    """Wrap a UUID (string or bytes) in a URN string"""
    import uuid

    try:
        b = b.decode("ascii")
    except AttributeError:
        pass
    return uuid.UUID(b).urn


def groupby(obj, keyfun):
    """Iterate through pprocs based on keys

    For example, if "stim" and "trial" are metadata on the trials, groupby(obj, lambda x: x["stim"]) will yield
    (stim0, [events_trial0, events_trial1]), (stim1, [events_trial0, events_trial1]), ...

    """
    import itertools

    evsorted = sorted(obj["pprox"], key=keyfun)
    return itertools.groupby(evsorted, key=keyfun)


def validate(obj):
    """Validates object against pprox schema"""
    import jsonschema
    import requests

    pass


def trial_iterator(pprox):
    """Iterates through trials in pprox, yielding (index, trial)"""
    for i, trial in enumerate(pprox["pprox"]):
        # annotate with sampling rate
        if "sampling_rate" not in trial["recording"]:
            entry = trial["recording"]["entry"]
            sampling_rate = pprox["entry_metadata"][entry]["sampling_rate"]
            trial["recording"]["sampling_rate"] = sampling_rate
        yield i, trial


def aggregate_events(pprox):
    """Aggregate all the events in a pprox into a single array, using the offset field to adjust times"""
    import numpy as np

    all_events = [
        np.asarray(trial["events"]) + trial["offset"] for trial in pprox["pprox"]
    ]
    return np.concatenate(all_events)


def combine_recordings(*pprox):
    """Combine events from multiple pprox objects into a single object, matching based on trial number"""
    pass


async def split_trial(trial, split_fun):
    """Split a trial into multiple intervals.

    split_fun: a coroutine that takes the name of the stimulus and returns a
    list of dictionaries, one per split. Each dict needs to have at least three
    fields: `stim_begin`, the time when the split begins (relative to the start
    of the stimulus; `stim_end`, the time when the split ends; and `name`, the
    name that should be given to the split. Any additional fields will be stored
    as metadata on the new trials. Alternatively, `split_fun` can return a
    pandas dataframe. Make sure the metadata fields are valid Python identifiers.

    Returns a pandas DataFrame with one row per split. Each split contains the
    events between the start of the (split) stimulus and the start of the next
    (split) stimulus. If there are no events in the interval, the `events` field
    will be nan. The times are referenced to the start of the stimulus in that
    split (t=0). The stim_end field indicates when the stimulus ended (relative
    to start of the stimulus), and the interval_end field indicates when the
    interval ended (relative to start of the stimulus). interval_end will be
    greater than stimulus_end when there is a gap between the end of one
    stimulus and the subsequent one. The last split is post-padded by the
    average of the gaps between all the other stimuli.

    Note that the splits are NOT pre-padded. t=0 corresponds to the beginning of
    the interval and the stimulus.

    """
    import pandas as pd

    stimulus = trial["stimulus"]
    stim_on = stimulus["interval"][0]
    splits = await split_fun(stimulus["name"])
    # calculate average gap between splits and use this to determine when
    # the last split should end
    gaps = splits.stim_begin.shift(-1) - splits.stim_end
    gaps.iloc[-1] = gaps.mean()
    splits["interval_end"] = splits.stim_end + gaps
    last_split_end = splits.interval_end.iloc[-1]
    spikes = pd.Series(trial["events"], dtype="f") - stim_on
    spikes = spikes[spikes < last_split_end]
    # this expression uses searchsorted to assign each spike to a split,
    # then groups the spikes by split and merges this with the table of splits
    df = splits.join(
        spikes.groupby(
            splits.stim_begin.searchsorted(spikes, side="left") - 1, group_keys=False
        )
        .apply(lambda x: x.to_numpy())
        .rename("events")
    )
    df["offset"] = trial["offset"] + df.stim_begin + stim_on
    df["events"] -= df.stim_begin
    df["stim_end"] -= df.stim_begin
    df["interval_end"] -= df.stim_begin
    df["source_trial"] = trial["index"]
    return df.drop(columns=["stim_begin"]).rename_axis(index="interval").reset_index()
